from collections import Counter
from copy import copy
from distutils.util import strtobool
from json import dumps, loads
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional

from django.urls import NoReverseMatch, reverse
from django.utils.safestring import mark_safe
from django.utils.text import slugify

from NEMO.exceptions import RequiredUnansweredQuestionsException
from NEMO.models import Consumable, ToolUsageCounter
from NEMO.views.consumables import make_withdrawal

dynamic_form_logger = getLogger(__name__)


class PostUsageQuestion:
	question_type = "Question"

	number_type = "number"
	float_type = "float"
	text_type = "textbox"
	textarea_type = "textarea"
	radio_type = "radio"
	dropdown_type = "dropdown"
	group_type = "group"
	question_types = [number_type, float_type, text_type, textarea_type, radio_type, dropdown_type, group_type]

	required_span = '<span style="color:red">*</span>'

	def __init__(self, properties: Dict, index: int = None):
		self.properties = properties
		self.name = self._init_property("name")
		self.title = self._init_property("title")
		self.title_html = self._init_property("title_html")
		self.help = self._init_property("help")
		self.type = self._init_property("type")
		self.max_width = self._init_property("max-width")
		self.maxlength = self._init_property("maxlength")
		self.placeholder = self._init_property("placeholder")
		self.prefix = self._init_property("prefix")
		self.suffix = self._init_property("suffix")
		self.pattern = self._init_property("pattern")
		self.min = self._init_property("min")
		self.max = self._init_property("max")
		self.precision = self._init_property("precision")
		self.step = self._init_property("step")
		self.rows = self._init_property("rows")
		self.consumable = self._init_property("consumable")
		self.required = self._init_property("required", True)
		self.default_choice = self._init_property("default_choice")
		self.choices = self._init_property("choices")
		self.group_add_button_name = self._init_property("group_add_button_name") or "Add"
		self.index = index
		if index and not isinstance(self, PostUsageGroupQuestion):
			self.name = f"{self.name}_{index}"
		# form_name is used in forms and extraction to avoid potential conflicts with other data
		self.form_name = f"df_{self.name}"
		pass

	def _init_property(self, prop: str, boolean: bool = False) -> Any:
		if boolean:
			return True if prop in self.properties and self.properties[prop] is True else False
		else:
			return self.properties[prop] if prop in self.properties else None

	def validate(self):
		self.validate_property_exists("name")
		self.validate_property_exists("title")
		self.validate_property_exists("type")

	def render(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		return self.render_element(virtual_inputs, group_question_url, group_item_id) + self.render_script(
			virtual_inputs, group_question_url, group_item_id
		)

	def render_as_text(self) -> str:
		result = f"{self.title}\n"
		result += "<strong>your answer</strong>"
		if self.choices:
			result += " (possible choices: " + "|".join(self.choices) + ")"
		return result

	def render_element(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		return ""

	def render_script(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		return ""

	def extract(self, request) -> Dict:
		answered_question = copy(self.properties)
		user_input = request.POST.get(self.form_name)
		if user_input:
			answered_question["user_input"] = user_input
		return answered_question

	def validate_property_exists(self, prop: str):
		try:
			self.properties[prop]
		except KeyError:
			raise Exception(f"{self.question_type} requires property '{prop}' to be defined")

	@staticmethod
	def load_questions(questions: Optional[List[Dict]], index: int = None):
		questions_to_load = questions or []
		constructor = {
			PostUsageQuestion.number_type: PostUsageNumberFieldQuestion,
			PostUsageQuestion.float_type: PostUsageFloatFieldQuestion,
			PostUsageQuestion.text_type: PostUsageTextFieldQuestion,
			PostUsageQuestion.textarea_type: PostUsageTextAreaFieldQuestion,
			PostUsageQuestion.radio_type: PostUsageRadioQuestion,
			PostUsageQuestion.dropdown_type: PostUsageDropdownQuestion,
			PostUsageQuestion.group_type: PostUsageGroupQuestion,
		}
		post_usage_questions: List[PostUsageQuestion] = []
		for question in questions_to_load:
			post_usage_questions.append(constructor.get(question["type"], PostUsageQuestion)(question, index))
		return post_usage_questions


class PostUsageRadioQuestion(PostUsageQuestion):
	question_type = "Question of type radio"

	def validate(self):
		super().validate()
		self.validate_property_exists("choices")

	def render_element(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		title = self.title_html or self.title
		result = f'<div class="form-group"><div style="white-space: pre-wrap">{title}{self.required_span if self.required else ""}</div>'
		for choice in self.choices:
			result += '<div class="radio">'
			required = "required" if self.required else ""
			is_default_choice = "checked" if self.default_choice and self.default_choice == choice else ""
			result += f'<label><input type="radio" name="{self.form_name}" value="{choice}" {required} {is_default_choice}>{choice}</label>'
			result += "</div>"
		result += "</div>"
		return result


class PostUsageDropdownQuestion(PostUsageQuestion):
	question_type = "Question of type dropdown"

	def validate(self):
		super().validate()
		self.validate_property_exists("max-width")
		self.validate_property_exists("choices")

	def render_element(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		title = self.title_html or self.title
		max_width = f"max-width:{self.max_width}px"
		result = f'<div class="form-group"><div style="white-space: pre-wrap">{title}{self.required_span if self.required else ""}</div>'
		required = "required" if self.required else ""
		result += (
			f'<select name="{self.form_name}" {required} style="margin-top: 5px;{max_width}" class="form-control">'
		)
		blank_disabled = 'disabled="disabled"' if required else ""
		placeholder = self.placeholder if self.placeholder else "Select an option"
		result += f'<option {blank_disabled} selected="selected" value="">{placeholder}</option>'
		for choice in self.choices:
			is_default_choice = "selected" if self.default_choice and self.default_choice == choice else ""
			result += f'<option value="{choice}" {is_default_choice}>{choice}</option>'
		result += "</select>"
		if self.help:
			result += f'<div style="font-size:smaller;color:#999;{max_width}">{self.help}</div>'
		result += "</div>"
		return result


class PostUsageTextFieldQuestion(PostUsageQuestion):
	question_type = "Question of type text"

	def validate(self):
		super().validate()
		self.validate_property_exists("max-width")

	def render_element(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		title = self.title_html or self.title
		max_width = f"max-width:{self.max_width}px"
		result = '<div class="form-group">'
		result += f'<label for="{self.form_name}" style="white-space: pre-wrap">{title}{self.required_span if self.required else ""}</label>'
		input_group_required = True if self.prefix or self.suffix else False
		if input_group_required:
			result += f'<div class="input-group" style="{max_width}">'
		if self.prefix:
			result += f'<span class="input-group-addon">{self.prefix}</span>'
		required = "required" if self.required else ""
		pattern = f'pattern="{self.pattern}"' if self.pattern else ""
		placeholder = f'placeholder="{self.placeholder}"' if self.placeholder else ""
		result += self.render_input(required, pattern, placeholder)
		if self.suffix:
			result += f'<span class="input-group-addon">{self.suffix}</span>'
		if input_group_required:
			result += "</div>"
		if self.help:
			result += f'<div style="font-size:smaller;color:#999;{max_width}">{self.help}</div>'
		result += "</div>"
		return result

	def render_input(self, required: str, pattern: str, placeholder: str) -> str:
		maxlength = f'maxlength="{self.maxlength}"' if self.maxlength else ""
		return f'<input type="text" class="form-control" id="{self.form_name}" name="{self.form_name}" {maxlength} {placeholder} {pattern} {required} style="max-width:{self.max_width}px" spellcheck="false" autocapitalize="off" autocomplete="off" autocorrect="off">'

	def render_script(self, virtual_inputs: bool, group_question_url: str, item_id: int) -> str:
		if virtual_inputs:
			return f"<script>$('#{self.form_name}').keyboard();</script>"
		return super().render_script(virtual_inputs, group_question_url, item_id)

	def render_as_text(self) -> str:
		result = f"{self.title}\n"
		if self.prefix:
			result += self.prefix
		result += "<strong>your answer</strong>"
		if self.suffix:
			result += f" {self.suffix}"
		return result


class PostUsageTextAreaFieldQuestion(PostUsageTextFieldQuestion):
	question_type = "Question of type textarea"

	def render_input(self, required: str, pattern: str, placeholder: str) -> str:
		rows = f'rows="{str(self.rows)}"' if self.rows else ""
		return f'<textarea class="form-control" id="{self.form_name}" name="{self.form_name}" {rows} {placeholder} {required} style="max-width:{self.max_width}px;height:inherit" spellcheck="false" autocapitalize="off" autocomplete="off" autocorrect="off"></textarea>'


class PostUsageNumberFieldQuestion(PostUsageTextFieldQuestion):
	question_type = "Question of type number"

	def render_input(self, required: str, pattern: str, placeholder: str) -> str:
		minimum = f'min="{self.min}"' if self.min else ""
		maximum = f'max="{self.max}"' if self.max else ""
		step = f'step="{self.step}"' if self.step else ""
		return f'<input type="number" class="form-control" id="{self.form_name}" name="{self.form_name}" {placeholder} {pattern} {minimum} {maximum} {step} {required} style="max-width:{self.max_width}px" spellcheck="false" autocapitalize="off" autocomplete="off" autocorrect="off">'

	def render_script(self, virtual_inputs: bool, group_question_url: str, item_id: int) -> str:
		if virtual_inputs:
			return f"<script>$('#{self.form_name}').numpad({{'readonly': false, 'hidePlusMinusButton': true, 'hideDecimalButton': true}});</script>"
		return super().render_script(virtual_inputs, group_question_url, item_id)

	def render_as_text(self) -> str:
		result = super().render_as_text()
		if self.min or self.max:
			result += " ("
			if self.min:
				result += f"min: {self.min}"
			if self.max:
				if self.min:
					result += ", "
				result += f"max: {self.min}"
			result += ")"
		return result


class PostUsageFloatFieldQuestion(PostUsageTextFieldQuestion):
	question_type = "Question of type float"

	def render_input(self, required: str, pattern: str, placeholder: str) -> str:
		precision = self.precision if self.precision else 2
		pattern = f'pattern="^\s*(?=.*[0-9])\d*(?:\.\d{"{1," + str(precision) + "}"})?\s*$"'
		return f'<input type="text" class="form-control" id="{self.form_name}" name="{self.form_name}" {placeholder} {pattern} {required} style="max-width:{self.max_width}px" spellcheck="false" autocapitalize="off" autocomplete="off" autocorrect="off">'

	def render_script(self, virtual_inputs: bool, group_question_url: str, item_id: int) -> str:
		if virtual_inputs:
			return f"<script>$('#{self.form_name}').numpad({{'readonly': false, 'hidePlusMinusButton': true, 'hideDecimalButton': false}});</script>"
		return super().render_script(virtual_inputs, group_question_url, item_id)


class PostUsageGroupQuestion(PostUsageQuestion):
	question_type = "Question of type group"

	def __init__(self, properties: Dict, index: int = None):
		super().__init__(properties, index)
		self.max_number = self._init_property("max_number")
		# we need a safe group name to use in js function and variable names
		self.group_name = slugify(self.name).replace("-", "_")
		self.sub_questions: List[PostUsageQuestion] = PostUsageQuestion.load_questions(
			self._init_property("questions"), index
		)

	def validate(self):
		super().validate()
		self.validate_property_exists("questions")
		self.validate_property_exists("max_number")
		for sub_question in self.sub_questions:
			sub_question.validate()

	def render_element(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		title = self.title_html or self.title
		result = f'<div class="form-group"><div style="white-space: pre-wrap">{title}</div></div>'
		result += f'<div id="{self.group_name}_container">'
		result += self.render_group_question(virtual_inputs, group_question_url, group_item_id)
		result += "</div>"
		result += "</div>"
		result += '<div class="form-group">'
		result += f'<button id="{self.group_name}_add_button" type="button" onclick="add_question_{self.group_name}()">{self.group_add_button_name}</button>'
		result += "</div>"
		return result

	def render_group_question(self, virtual_inputs: bool, group_question_url: str, group_item_id: int) -> str:
		result = ""
		result += f'<div class="{self.group_name}_question" style="padding-top: 10px; padding-bottom: 10px; border-top: 1px solid lightgray">'
		for sub_question in self.sub_questions:
			result += sub_question.render(virtual_inputs, group_question_url, group_item_id)
		if self.index:
			result += f'<button type="button" onclick="remove_question_{self.group_name}(this);">Remove</button>'
		return result

	def render_script(self, virtual_inputs: bool, group_question_url: str, item_id: int) -> str:
		return f"""
		<script>
			if (!$) {{ $ = django.jQuery; }}
			var {self.group_name}_question_index=1;
			function update_add_button_{self.group_name}()
			{{
				if ($(".{self.group_name}_question").length < {self.max_number})
				{{
					$("#{self.group_name}_add_button").show();
				}}
				else
				{{
					$("#{self.group_name}_add_button").hide();
				}}
			}};
			function remove_question_{self.group_name}(element)
			{{
				$(element).parents(".{self.group_name}_question").remove();
				$("body").trigger("question-group-changed");
				update_add_button_{self.group_name}();
			}}
			function add_question_{self.group_name}()
			{{
				$.ajax({{ type: "GET", url: "{reverse(group_question_url, args=[item_id, self.group_name])}?virtual_inputs={virtual_inputs}&index="+{self.group_name}_question_index, success : function(response)
				{{
					{self.group_name}_question_index ++;
					$("#{self.group_name}_container").append(response);
					$("body").trigger("question-group-changed");
					update_add_button_{self.group_name}();
				}}
				}});
			}}
		</script>"""

	def render_as_text(self) -> str:
		result = f"{self.title}\n"
		for question in self.sub_questions:
			result += question.render_as_text()
		return result

	def extract(self, request) -> Dict:
		# For group question, we also have to look for answers submitted with a numbered suffix (i.e. question, question_1, question_2 etc.)
		# The result of the extraction will be a dictionary, with the keys being the group number, and the values the user inputs for all questions of the group.
		sub_results = copy(self.properties)
		user_inputs = {}
		question_form_names = [sub_question.form_name for sub_question in self.sub_questions]
		for key, value in request.POST.items():
			if key in question_form_names:
				for sub_question in self.sub_questions:
					if key == sub_question.form_name:
						user_inputs.setdefault(0, {})
						user_inputs[0][sub_question.name] = value
			else:
				for sub_question in self.sub_questions:
					name = sub_question.form_name
					if key.startswith(name + "_"):
						index = int(key.rsplit("_", 1)[1])
						user_inputs.setdefault(index, {})
						user_inputs[index][sub_question.name] = value
		sub_results["user_input"] = user_inputs
		return sub_results


class DynamicForm:
	def __init__(self, questions):
		self.untreated_questions = []
		self.questions = []
		if questions:
			self.untreated_questions = loads(questions)
			self.questions: List[PostUsageQuestion] = PostUsageQuestion.load_questions(self.untreated_questions)

	def render(self, group_question_url: str, group_item_id: int, virtual_inputs: bool = False):
		result = ""
		for question in self.questions:
			result += question.render(virtual_inputs, group_question_url, group_item_id)
		return mark_safe(result)

	def validate(self, group_question_url: str, group_item_id: int):
		# We need to validate the raw json for types
		for question in self.untreated_questions:
			if question["type"] not in PostUsageQuestion.question_types:
				raise Exception(f"type has to be one of {', '.join(PostUsageQuestion.question_types)}")
			if question["type"] == PostUsageQuestion.group_type and "questions" in question:
				for sub_question in question["questions"]:
					if sub_question["type"] not in PostUsageQuestion.question_types:
						raise Exception(f"type has to be one of {', '.join(PostUsageQuestion.question_types)}")
		for question in self.questions:
			question.validate()
		# Test the rendering, but catch reverse exception if this the item doesn't have an id yet (when creating it the first time)
		try:
			self.render(group_question_url, group_item_id)
		except NoReverseMatch:
			if group_item_id:
				raise
			pass
		# Check for duplicate names
		names = []
		for question in self.questions:
			names.append(question.name)
			if isinstance(question, PostUsageGroupQuestion):
				for sub_question in question.sub_questions:
					names.append(sub_question.name)
		duplicate_names = [k for k, v in Counter(names).items() if v > 1]
		if duplicate_names:
			raise Exception(f"Question names need to be unique. Duplicates were found: {duplicate_names}")
		# Check that consumable exists and is linked to a number question
		for question in self.questions:
			if question.consumable:
				if not isinstance(question, PostUsageNumberFieldQuestion):
					raise Exception("Consumable withdrawals can only be used in questions of type number")
				else:
					try:
						Consumable.objects.get(name=question.consumable)
					except Consumable.DoesNotExist:
						raise Exception(
							f"Consumable with name '{question.consumable}' could not be found. Make sure the names match."
						)
			if isinstance(question, PostUsageGroupQuestion):
				for sub_question in question.sub_questions:
					if sub_question.consumable:
						raise Exception("Consumable withdrawals cannot be used in group questions")

	def extract(self, request) -> str:
		results = {}
		required_unanswered_questions = []
		for question in self.questions:
			results[question.name] = question.extract(request)
			required_unanswered_questions.extend(self._check_for_required_unanswered_questions(results, question))
		run_data = dumps(results, indent="\t") if len(results) else ""
		if required_unanswered_questions:
			raise RequiredUnansweredQuestionsException(run_data, required_unanswered_questions)
		return run_data

	def _check_for_required_unanswered_questions(self, results: Dict, question: PostUsageQuestion) -> Optional[List[PostUsageQuestion]]:
		# This method will check for required unanswered questions and if some are found, will fill them with blank and return them
		required_unanswered_questions = []
		user_input = results[question.name].get("user_input")
		if not isinstance(question, PostUsageGroupQuestion) and question.required and not user_input:
			results[question.name]["user_input"] = ""
			required_unanswered_questions.append(question)
		elif isinstance(question, PostUsageGroupQuestion):
			blank_user_input = {0: {}}
			for sub_question in question.sub_questions:
				if sub_question.required and (not user_input or not user_input.get(0, {}).get(sub_question.name)):
					blank_user_input[0][sub_question.name] = ""
					required_unanswered_questions.append(sub_question)
			if required_unanswered_questions:
				results[question.name]["user_input"] = blank_user_input
		return required_unanswered_questions

	def filter_questions(self, function: Callable[[PostUsageQuestion], bool]) -> List[PostUsageQuestion]:
		results = []
		for question in self.questions:
			if function(question):
				results.append(question)
			elif isinstance(question, PostUsageGroupQuestion):
				for sub_question in question.sub_questions:
					if function(sub_question):
						results.append(sub_question)
		return results

	def charge_for_consumables(self, customer, merchant, project, run_data: str, request=None):
		try:
			run_data_json = loads(run_data)
		except Exception as error:
			dynamic_form_logger.debug(error)
			return
		for question in self.questions:
			if question.consumable:
				try:
					consumable = Consumable.objects.get(name=question.consumable)
					quantity = 0
					if isinstance(question, PostUsageNumberFieldQuestion):
						if (
								question.name in run_data_json
								and "user_input" in run_data_json[question.name]
								and run_data_json[question.name]["user_input"]
						):
							quantity = int(run_data_json[question.name]["user_input"])
					if quantity > 0:
						make_withdrawal(
							consumable_id=consumable.id,
							customer_id=customer.id,
							merchant=merchant,
							quantity=quantity,
							project_id=project.id,
							request=request,
						)
				except Exception as e:
					dynamic_form_logger.warning(
						f"Could not withdraw consumable: '{question.consumable}' with quantity: '{run_data_json[question.name]}' for customer: '{customer}' by merchant: '{merchant}' for project: '{project}'",
						e,
					)
					pass

	def update_tool_counters(self, run_data: str, tool_id: int):
		# This function increments all counters associated with the given tool
		try:
			run_data_json = loads(run_data)
		except Exception as error:
			dynamic_form_logger.debug(error)
			return
		active_counters = ToolUsageCounter.objects.filter(is_active=True, tool_id=tool_id)
		for counter in active_counters:
			additional_value = 0
			for question in self.questions:
				if isinstance(question, PostUsageNumberFieldQuestion) or isinstance(
						question, PostUsageFloatFieldQuestion
				):
					if (
							question.name == counter.tool_usage_question
							and question.name in run_data_json
							and "user_input" in run_data_json[question.name]
							and run_data_json[question.name]["user_input"]
					):
						additional_value = float(run_data_json[question.name]["user_input"])
				elif isinstance(question, PostUsageGroupQuestion):
					for sub_question in question.sub_questions:
						if isinstance(sub_question, PostUsageNumberFieldQuestion) or isinstance(
								sub_question, PostUsageFloatFieldQuestion
						):
							if (
									sub_question.name == counter.tool_usage_question
									and question.name in run_data_json
									and "user_input" in run_data_json[question.name]
							):
								for user_input in run_data_json[question.name]["user_input"].values():
									if sub_question.name in user_input and user_input[sub_question.name]:
										additional_value += float(user_input[sub_question.name])
			if additional_value:
				counter.value += additional_value
				counter.save()


def render_group_questions(request, questions, group_question_url, group_item_id, group_name) -> str:
	question_index = request.GET["index"]
	virtual_inputs = bool(strtobool((request.GET["virtual_inputs"])))
	if questions:
		for question in PostUsageQuestion.load_questions(loads(questions), question_index):
			if isinstance(question, PostUsageGroupQuestion) and question.group_name == group_name:
				return question.render_group_question(virtual_inputs, group_question_url, group_item_id)
	return ""



# Generated by Django 2.2.13 on 2020-08-04 20:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0021_version_3_1_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='area',
            name='count_service_personnel_in_occupancy',
            field=models.BooleanField(default=True, help_text='Indicates that service personnel will count towards maximum capacity.'),
        ),
        migrations.AddField(
            model_name='user',
            name='is_service_personnel',
            field=models.BooleanField(default=False, help_text='Designates this user as service personnel. Service personnel can operate qualified tools without a reservation even when they are shutdown or during an outage and can access authorized areas without a reservation.', verbose_name='service personnel'),
        ),
        migrations.AddField(
            model_name='consumable',
            name='reminder_threshold_reached',
            field=models.BooleanField(default=False),
        ),
    ]



from django.http import HttpResponseBadRequest
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import staff_member_required
from NEMO.exceptions import ProjectChargeException
from NEMO.models import Area, AreaAccessRecord, Project, StaffCharge, UsageEvent, User
from NEMO.views.area_access import load_areas_for_use_in_template
from NEMO.views.policy import check_billing_to_project


@staff_member_required
@require_GET
def staff_charges(request):
	staff_member:User = request.user
	staff_charge: StaffCharge = staff_member.get_staff_charge()
	dictionary = dict()
	if staff_charge:
		try:
			dictionary['staff_charge'] = staff_charge
			# Create dictionary of charges for time, tool and areas
			charges = [{'type': 'Start time charge', 'start': staff_charge.start, 'end': staff_charge.end}]
			for area_charge in AreaAccessRecord.objects.filter(staff_charge_id=staff_charge.id):
				charges.append({'type': area_charge.area.name + ' access', 'start': area_charge.start, 'end': area_charge.end, 'class': 'primary-highlight'})
			for tool_charge in UsageEvent.objects.filter(operator=staff_member, user=staff_charge.customer, start__gt=staff_charge.start):
				charges.append({'type': tool_charge.tool.name + ' usage', 'start': tool_charge.start, 'end': tool_charge.end, 'class': 'warning-highlight'})
			charges.sort(key=lambda x: x['start'], reverse=True)
			dictionary['charges'] = charges

			area_access_record = AreaAccessRecord.objects.get(staff_charge=staff_charge.id, end=None)
			dictionary['area'] = area_access_record.area
			return render(request, 'staff_charges/end_area_charge.html', dictionary)
		except AreaAccessRecord.DoesNotExist:
			dictionary['user_accessible_areas'], dictionary['areas'] = load_areas_for_use_in_template(staff_member)
			return render(request, 'staff_charges/change_status.html', dictionary)
	error = None
	customer = None
	try:
		customer = User.objects.get(id=request.GET['customer'])
	except:
		pass
	if customer:
		if customer.active_project_count() > 0:
			dictionary['customer'] = customer
			return render(request, 'staff_charges/choose_project.html', dictionary)
		else:
			error = str(customer) + ' does not have any active projects. You cannot bill staff time to this user.'
	users = User.objects.filter(is_active=True).exclude(id=request.user.id)
	dictionary['users'] = users
	dictionary['error'] = error
	return render(request, 'staff_charges/new_staff_charge.html', dictionary)


@staff_member_required
@require_POST
def begin_staff_charge(request):
	if request.user.charging_staff_time():
		return HttpResponseBadRequest('You cannot create a new staff charge when one is already in progress.')
	charge = StaffCharge()
	charge.customer = User.objects.get(id=request.POST['customer'])
	charge.project = Project.objects.get(id=request.POST['project'])
	# Check if we are allowed to bill to project
	try:
		check_billing_to_project(charge.project, charge.customer, charge)
	except ProjectChargeException as e:
		return HttpResponseBadRequest(e.msg)
	charge.staff_member = request.user
	charge.save()
	return redirect(reverse('staff_charges'))


@staff_member_required
@require_POST
def end_staff_charge(request):
	if not request.user.charging_staff_time():
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot end it.')
	charge = request.user.get_staff_charge()
	charge.end = timezone.now()
	charge.save()
	try:
		area_access = AreaAccessRecord.objects.get(staff_charge=charge, end=None)
		area_access.end = timezone.now()
		area_access.save()
	except AreaAccessRecord.DoesNotExist:
		pass
	return redirect(reverse('staff_charges'))


@staff_member_required
@require_POST
def begin_staff_area_charge(request):
	charge = request.user.get_staff_charge()
	if not charge:
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot begin an area access charge.')
	if AreaAccessRecord.objects.filter(staff_charge=charge, end=None).count() > 0:
		return HttpResponseBadRequest('You cannot create an area access charge when one is already in progress.')
	try:
		area = Area.objects.get(id=request.POST['area'])
		check_billing_to_project(charge.project, charge.customer, area)
	except ProjectChargeException as e:
		return HttpResponseBadRequest(e.msg)
	except:
		return HttpResponseBadRequest('Invalid area')
	area_access = AreaAccessRecord()
	area_access.area = area
	area_access.staff_charge = charge
	area_access.customer = charge.customer
	area_access.project = charge.project
	area_access.save()
	return redirect(reverse('staff_charges'))


@staff_member_required
@require_POST
def end_staff_area_charge(request):
	charge = request.user.get_staff_charge()
	if not charge:
		return HttpResponseBadRequest('You do not have a staff charge in progress, so you cannot end area access.')
	area_access = AreaAccessRecord.objects.get(staff_charge=charge, end=None)
	area_access.end = timezone.now()
	area_access.save()
	return redirect(reverse('staff_charges'))



from django.contrib.auth.decorators import login_required
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO.decorators import staff_member_required
from NEMO.forms import AccountForm, ProjectForm
from NEMO.models import Account, ActivityHistory, MembershipHistory, Project, User
from NEMO.views.pagination import SortedPaginator


@staff_member_required
@require_GET
def accounts_and_projects(request):
	all_accounts = Account.objects.all().order_by("name")

	page = SortedPaginator(all_accounts, request, order_by="name").get_current_page()

	dictionary = {"page": page, "accounts_and_projects": set(Account.objects.all()) | set(Project.objects.all())}
	return render(request, "accounts_and_projects/accounts_and_projects.html", dictionary)


@staff_member_required
@require_GET
def select_accounts_and_projects(request, kind=None, identifier=None):
	selected_project = None
	try:
		if kind == "project":
			selected_project = Project.objects.get(id=identifier)
			account = selected_project.account
		elif kind == "account":
			account = Account.objects.get(id=identifier)
		else:
			account = None
	except:
		account = None
	dictionary = {
		"account": account,
		"selected_project": selected_project,
		"accounts_and_projects": set(Account.objects.all()) | set(Project.objects.all()),
		"users": User.objects.all(),
	}
	return render(request, "accounts_and_projects/account_and_projects.html", dictionary)


@staff_member_required
@require_POST
def toggle_active(request, kind, identifier):
	if kind == "account":
		entity = get_object_or_404(Account, id=identifier)
	elif kind == "project":
		entity = get_object_or_404(Project, id=identifier)
	else:
		return HttpResponseBadRequest("Invalid entity for active toggle request.")
	entity.active = not entity.active
	entity.save()
	history = ActivityHistory()
	history.authorizer = request.user
	history.action = entity.active
	history.content_object = entity
	history.save()
	return redirect(request.META.get("HTTP_REFERER", "accounts_and_projects"))


@staff_member_required
@require_http_methods(["GET", "POST"])
def create_project(request):
	form = ProjectForm(request.POST or None)
	dictionary = {"account_list": Account.objects.all(), "user_list": User.objects.filter(is_active=True), "form": form}
	if request.method == "GET":
		return render(request, "accounts_and_projects/create_project.html", dictionary)
	if not form.is_valid():
		return render(request, "accounts_and_projects/create_project.html", dictionary)
	project = form.save()
	account_history = MembershipHistory()
	account_history.authorizer = request.user
	account_history.action = MembershipHistory.Action.ADDED
	account_history.child_content_object = project
	account_history.parent_content_object = project.account
	account_history.save()
	project_history = ActivityHistory()
	project_history.authorizer = request.user
	project_history.action = project.active
	project_history.content_object = project
	project_history.save()
	return redirect("project", project.id)


@staff_member_required
@require_http_methods(["GET", "POST"])
def create_account(request):
	form = AccountForm(request.POST or None)
	dictionary = {"form": form}
	if request.method == "GET":
		return render(request, "accounts_and_projects/create_account.html", dictionary)
	if not form.is_valid():
		return render(request, "accounts_and_projects/create_account.html", dictionary)
	account = form.save()
	history = ActivityHistory()
	history.authorizer = request.user
	history.action = account.active
	history.content_object = account
	history.save()
	return redirect("account", account.id)


@staff_member_required
@require_POST
def remove_user_from_project(request):
	user = get_object_or_404(User, id=request.POST["user_id"])
	project = get_object_or_404(Project, id=request.POST["project_id"])
	if project.user_set.filter(id=user.id).exists():
		history = MembershipHistory()
		history.action = MembershipHistory.Action.REMOVED
		history.authorizer = request.user
		history.parent_content_object = project
		history.child_content_object = user
		history.save()
		project.user_set.remove(user)
	dictionary = {"users": project.user_set.all(), "project": project}
	return render(request, "accounts_and_projects/users_for_project.html", dictionary)


@staff_member_required
@require_POST
def add_user_to_project(request):
	user = get_object_or_404(User, id=request.POST["user_id"])
	project = get_object_or_404(Project, id=request.POST["project_id"])
	if user not in project.user_set.all():
		history = MembershipHistory()
		history.action = MembershipHistory.Action.ADDED
		history.authorizer = request.user
		history.parent_content_object = project
		history.child_content_object = user
		history.save()
		project.user_set.add(user)
	dictionary = {"users": project.user_set.all(), "project": project}
	return render(request, "accounts_and_projects/users_for_project.html", dictionary)


@login_required
@require_GET
def projects(request):
	user: User = request.user
	dictionary = {"managed_projects": user.managed_projects.all()}
	return render(request, "accounts_and_projects/projects.html", dictionary)



from setuptools import find_packages, setup

setup(
	name='NEMO',
	version='3.13.3',
	python_requires='>=3.6',
	packages=find_packages(exclude=['NEMO.tests','NEMO.tests.*']),
	include_package_data=True,
	url='https://github.com/usnistgov/NEMO',
	license='Public domain',
	author='Center for Nanoscale Science and Technology',
	author_email='CNSTapplications@nist.gov',
	description='NEMO is a laboratory logistics web application. Use it to schedule reservations, control tool access, track maintenance issues, and more.',
	long_description='Find out more about NEMO on the GitHub project page https://github.com/usnistgov/NEMO',
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Environment :: Web Environment',
		'Framework :: Django',
		'Intended Audience :: Science/Research',
		'Intended Audience :: System Administrators',
		'License :: Public Domain',
		'Natural Language :: English',
		'Operating System :: OS Independent',
		'Programming Language :: Python :: 3.6',
	],
	install_requires=[
		'cryptography==3.4.7',
		'Django==2.2.24',
		'django-filter==2.4.0',
		'drf-renderer-xlsx==0.4.1',
		'djangorestframework==3.12.4',
		'drf-flex-fields==0.9.3',
		'ldap3==2.9',
		'python-dateutil==2.8.1',
		'requests==2.25.1',
		'Pillow==8.3.2',
		'django-mptt==0.12.0',
	],
	entry_points={
		'console_scripts': ['nemo=NEMO.provisioning:entry_point'],
	},
)



from time import sleep

from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, render
from django.template.defaultfilters import linebreaksbr
from django.utils import timezone
from django.utils.formats import localize
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import postpone
from NEMO.exceptions import (
	InactiveUserError,
	MaximumCapacityReachedError,
	NoAccessiblePhysicalAccessUserError,
	NoActiveProjectsForUserError,
	NoPhysicalAccessUserError,
	PhysicalAccessExpiredUserError,
	ProjectChargeException,
	ReservationRequiredUserError,
	ScheduledOutageInProgressError,
	UnavailableResourcesUserError,
)
from NEMO.models import (BadgeReader, Door, PhysicalAccessLog, PhysicalAccessType, Project, UsageEvent, User)
from NEMO.views.area_access import log_in_user_to_area, log_out_user
from NEMO.views.customization import get_customization
from NEMO.views.policy import check_billing_to_project, check_policy_to_enter_any_area, check_policy_to_enter_this_area
from NEMO.views.tool_control import interlock_bypass_allowed


@login_required
@permission_required("NEMO.add_areaaccessrecord")
@require_GET
def welcome_screen(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	reader_id = request.GET.get("reader_id")
	badge_reader = BadgeReader.objects.get(id=reader_id) if reader_id else BadgeReader.default()
	return render(
		request, "area_access/welcome_screen.html", {"area": door.area, "door": door, "badge_reader": badge_reader}
	)


@login_required
@permission_required("NEMO.change_areaaccessrecord")
@require_GET
def farewell_screen(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	reader_id = request.GET.get("reader_id")
	badge_reader = BadgeReader.objects.get(id=reader_id) if reader_id else BadgeReader.default()
	return render(
		request, "area_access/farewell_screen.html", {"area": door.area, "door": door, "badge_reader": badge_reader}
	)


@login_required
@permission_required("NEMO.add_areaaccessrecord")
@require_POST
def login_to_area(request, door_id):
	door = get_object_or_404(Door, id=door_id)

	badge_number = request.POST.get("badge_number")
	bypass_interlock = request.POST.get("bypass", 'False') == 'True'
	if not badge_number:
		return render(request, "area_access/badge_not_found.html")
	try:
		user = User.objects.get(badge_number=badge_number)
	except User.DoesNotExist:
		return render(request, "area_access/badge_not_found.html")

	log = PhysicalAccessLog()
	log.user = user
	log.door = door
	log.time = timezone.now()
	log.result = PhysicalAccessType.DENY  # Assume the user does not have access

	facility_name = get_customization("facility_name")

	# Check policy for entering an area
	try:
		check_policy_to_enter_any_area(user=user)
	except InactiveUserError:
		log.details = "This user is not active, preventing them from entering any access controlled areas."
		log.save()
		return render(request, "area_access/inactive.html")

	except NoActiveProjectsForUserError:
		log.details = "The user has no active projects, preventing them from entering an access controlled area."
		log.save()
		return render(request, "area_access/no_active_projects.html")

	except PhysicalAccessExpiredUserError:
		log.details = "This user was blocked from this physical access level because their physical access has expired."
		log.save()
		message = f"Your physical access to the {facility_name} has expired. Have you completed your safety training within the last year? Please visit the User Office to renew your access."
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	except NoPhysicalAccessUserError:
		log.details = "This user does not belong to ANY physical access levels."
		log.save()
		message = f"You have not been granted physical access to any {facility_name} area. Please visit the User Office if you believe this is an error."
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	max_capacity_reached = False
	reservation_requirement_failed = False
	scheduled_outage_in_progress = False
	# Check policy to enter this area
	try:
		check_policy_to_enter_this_area(area=door.area, user=user)
	except NoAccessiblePhysicalAccessUserError as error:
		if error.access_exception:
			log.details = (
				f"The user was blocked from entering this area because of an exception: {error.access_exception.name}."
			)
			message = f"You do not have access to this area of the {facility_name} due to the following exception: {error.access_exception}. The exception ends on {localize(error.access_exception.end_time.astimezone(timezone.get_current_timezone()))}"
		else:
			log.details = (
				"This user is not assigned to a physical access level that allows access to this door at this time."
			)
			message = f"You do not have access to this area of the {facility_name} at this time. Please visit the User Office if you believe this is an error."
		log.save()
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	except UnavailableResourcesUserError as error:
		log.details = f"The user was blocked from entering this area because a required resource was unavailable [{', '.join(str(resource) for resource in error.resources)}]."
		log.save()
		return render(request, "area_access/resource_unavailable.html", {"unavailable_resources": error.resources})

	except MaximumCapacityReachedError as error:
		# deal with this error after checking if the user is already logged in
		max_capacity_reached = error

	except ScheduledOutageInProgressError as error:
		# deal with this error after checking if the user is already logged in
		scheduled_outage_in_progress = error

	except ReservationRequiredUserError:
		# deal with this error after checking if the user is already logged in
		reservation_requirement_failed = True

	current_area_access_record = user.area_access_record()
	if current_area_access_record and current_area_access_record.area == door.area:
		# No log entry necessary here because all validation checks passed.
		# The log entry is captured when the subsequent choice is made by the user.
		return render(
			request,
			"area_access/already_logged_in.html",
			{
				"area": door.area,
				"project": current_area_access_record.project,
				"badge_number": user.badge_number,
				"reservation_requirement_failed": reservation_requirement_failed,
				"max_capacity_reached": max_capacity_reached,
				"scheduled_outage_in_progress": scheduled_outage_in_progress,
			},
		)

	if scheduled_outage_in_progress:
		log.details = f"The user was blocked from entering this area because the {scheduled_outage_in_progress.area.name} has a scheduled outage in progress."
		log.save()
		message = (
			f"The {scheduled_outage_in_progress.area.name} is inaccessible because a scheduled outage is in progress."
		)
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	if max_capacity_reached:
		log.details = f"The user was blocked from entering this area because the {max_capacity_reached.area.name} has reached its maximum capacity of {max_capacity_reached.area.maximum_capacity} people at a time."
		log.save()
		message = f"The {max_capacity_reached.area.name} has reached its maximum capacity. Please wait for somebody to leave and try again."
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	if reservation_requirement_failed:
		log.details = f"The user was blocked from entering this area because the user does not have a current reservation for the {door.area}."
		log.save()
		message = "You do not have a current reservation for this area. Please make a reservation before trying to access this area."
		return render(request, "area_access/physical_access_denied.html", {"message": message})

	if user.active_project_count() >= 1:
		if user.active_project_count() == 1:
			project = user.active_projects()[0]
		else:
			project_id = request.POST.get("project_id")
			if not project_id:
				# No log entry necessary here because all validation checks passed, and the user must indicate which project
				# the wish to login under. The log entry is captured when the subsequent choice is made by the user.
				return render(request, "area_access/choose_project.html", {"area": door.area, "user": user})
			else:
				project = get_object_or_404(Project, id=project_id)
				try:
					check_billing_to_project(project, user, door.area)
				except ProjectChargeException as e:
					log.details = "The user attempted to bill the project named {} but got error: {}".format(project.name, e.msg)
					log.save()
					return render(request, "area_access/physical_access_denied.html", {"message": e.msg})

		log.result = PhysicalAccessType.ALLOW
		log.save()

		# Automatically log the user out of any previous area before logging them in to the new area.
		previous_area = None
		if user.in_area():
			previous_area = user.area_access_record().area
			log_out_user(user)

		# All policy checks passed so open the door for the user.
		if not door.interlock.unlock():
			if bypass_interlock and interlock_bypass_allowed(user):
				pass
			else:
				return interlock_error("Login", user)

		delay_lock_door(door.id)

		log_in_user_to_area(door.area, user, project)

		return render(
			request,
			"area_access/login_success.html",
			{"area": door.area, "name": user.first_name, "project": project, "previous_area": previous_area},
		)


@postpone
def delay_lock_door(door_id):
	door = Door.objects.get(id=door_id)
	sleep(8)
	door.interlock.lock()


@login_required
@permission_required("NEMO.change_areaaccessrecord")
@require_POST
def logout_of_area(request, door_id):
	badge_number = request.POST.get("badge_number")
	if not badge_number:
		return render(request, "area_access/badge_not_found.html")
	try:
		user = User.objects.get(badge_number=badge_number)
	except User.DoesNotExist:
		return render(request, "area_access/badge_not_found.html")
	record = user.area_access_record()
	if record:
		log_out_user(user)
		busy_tools = UsageEvent.objects.filter(end=None, user=user)
		staff_charge = user.get_staff_charge()
		if busy_tools or staff_charge:
			return render(
				request,
				"area_access/logout_warning.html",
				{
					"area": record.area,
					"name": user.first_name,
					"tools_in_use": busy_tools,
					"staff_charge": staff_charge,
				},
			)
		else:
			return render(request, "area_access/logout_success.html", {"area": record.area, "name": user.first_name})
	else:
		return render(request, "area_access/not_logged_in.html")


@login_required
@permission_required("NEMO.change_areaaccessrecord")
@require_POST
def open_door(request, door_id):
	door = get_object_or_404(Door, id=door_id)
	badge_number = request.POST.get("badge_number")
	if not badge_number:
		return render(request, "area_access/badge_not_found.html")
	try:
		user = User.objects.get(badge_number=badge_number)
	except User.DoesNotExist:
		return render(request, "area_access/badge_not_found.html")
	if user.area_access_record() and user.area_access_record().area == door.area:
		log = PhysicalAccessLog(
			user=user,
			door=door,
			time=timezone.now(),
			result=PhysicalAccessType.ALLOW,
			details="The user was permitted to enter this area, and already had an active area access record for this area.",
		)
		log.save()
		# If we cannot open the door, display message and let them try again or exit since there is nothing else to do (user is already logged in).
		if not door.interlock.unlock():
			return interlock_error(bypass_allowed=False)
		delay_lock_door(door.id)
		return render(request, "area_access/door_is_open.html")
	return render(request, "area_access/not_logged_in.html", {"area": door.area})


def interlock_error(action: str = None, user: User = None, bypass_allowed: bool = None):
	error_message = get_customization('door_interlock_failure_message')
	bypass_allowed = interlock_bypass_allowed(user) if bypass_allowed is None else bypass_allowed
	dictionary = {
		"message": linebreaksbr(error_message),
		"bypass_allowed": bypass_allowed,
		"action": action
	}
	return JsonResponse(dictionary, status=501)



# Generated by Django 2.2.10 on 2020-05-29 16:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0018_area_count_staff_in_occupancy'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='NEMO.UserType'),
        ),
    ]
